#!/usr/bin/perl -w

# Larry - your friendly music server bot

use IO::Socket;
use strict;

use POSIX qw(:errno_h);


# Global definitions

my $server = $ARGV[0] || "music";
my $port   = $ARGV[1] || 7778;
my $botname = $ARGV[2] || "kraken";
my $brain = $ARGV[3] || "$ENV{HOME}/usr/lib/puke/larrybrain.pl";

require $brain;

my $debug = 0;

# Global variables

my $sock;

# Various functions and stuff

sub logit
  {
    my ($msg) = @_;

    if ($debug <= 0)
      {
        if ($msg =~ /^\s*Debug:/i)
          {
            return;
          }
      }

    print STDERR "$msg\n";
  }

sub signon
  {
    logit "Larry: connecing to $server:$port";

    $sock = IO::Socket::INET->new(PeerAddr => $server,
				  PeerPort => $port,
				  Proto    => 'tcp')
      or die "The server is not up";

    logit "Larry: connected to server";
    sleep 1;
    flush $sock;
    larry_init($sock, $botname);
    logit "Larry: using name $botname";
    
    print $sock "$botname\n";
    

    flush $sock;

    my $signedon = 0;

    while (!$signedon)
      {

	my $line = &next_line();

	if (! defined $line && !$signedon)
	  {
	    $signedon = 2;
	    logit "Larry: eof";
	  }

	if ($line =~ /Hello, $botname/i)
	  {
	    logit "Larry: ooh, we stir. Looks like we're in business";
	    $signedon = 1;
	  }
      }

    if ($signedon == 2)
      {
	logit "Larry: server died";
	die "Server died";
      }
  }

# Main loop

my $io_buffer = '';
my @io_pending;

sub receive
  {
    my $buffer = '';
    my $rv = sysread($sock, $buffer, 1024);
    if (!defined($rv)) {
      if ($! == EAGAIN) {
        # would block
	warn "would block";
        return [];
      } else {
        die "sysread failed: $!";
      }
    } else {
      if ($rv) {
        # successfully read $rv bytes from HANDLE
        $io_buffer .= $buffer;
      } else {
        # End of file
        return undef;
      }
    }
    
    my @lines;
    
    # Chomp carridge returns
    $io_buffer =~ s/\r//g;
    # XXX fix me to use a regex in array context or split
    while ((my $pos = index $io_buffer, "\n") != -1) {
      push @lines, substr $io_buffer, 0, $pos;
      $io_buffer = substr $io_buffer, $pos+1;
      $io_buffer = '' unless defined $io_buffer;
    }
    
    return \@lines;
  }

sub next_line 
  {
    while (1) {
      if ($#io_pending != -1) {
        return shift @io_pending;
      }
    
      my $read = '';
      vec ($read, $sock->fileno, 1) = 1;

      my $nfound = select $read, undef, undef, undef;

      if (!(vec ($read, $sock->fileno, 1))) {
        next;
      }

      my $lines = &receive();
      return undef unless defined $lines;
      push @io_pending, @{$lines};
    }
  }


sub mainloop
  {
    while (1)
      {
        $_ = &next_line();
	
	last unless defined $_;
#        logit "[[$_]]";
	if (/^([!-~]+)\s*>([^\s].*)$/)
	  {
	    logit "Debug: $1 told us $2";
	    
	    eval { larry_tell($1, $2); }; logit "Error: $@" if ($@);
	  }
	if (/^([!-~]+)\s*:([^\s].*)$/)
	  {
	    logit "Debug: $1 said $2";
	    
	    eval { larry_say($1, $2); }; logit "Error: $@" if ($@);
	  }

	# Status message parsing

	elsif (/^\*\*\* Logged on today are:$/)
	  {
	    logit "Debug: start of list of users";
	    
	    eval { larry_user_start(); }; logit "Error: $@" if ($@);
	  }
	elsif (/^\*\*\* ([!-~]+) \((.*)\)( .*|)$/)
	  {
	    logit "Debug: user $1 $2 $3";

	    eval { larry_user($1, $2, $3); }; logit "Error: $@" if ($@);
	  }
	elsif (/^\*\*\* Matching tracks are:$/)
	  {
	    logit "Debug: Start of track list";
	    
	    eval { larry_track_start(); }; logit "Error: $@" if ($@);
	  }
	elsif (/^\*\*\* ([!-~]+)$/)
	  {
	    logit "Debug: Got track $1";
	    
	    eval { larry_track($1); }; logit "Error: $@" if ($@);
	  }
	elsif (/^\*\*\* (.+) has requested a server reinit$/)
	  {
	    logit "Debug: Get server reinit";
	    larry_reinit();
	  }
	elsif (/^\*\*\* Currently playing \'(.*)\' \(queued by ([!-~]+)\)$/)
	  {
	    logit "Debug: start of queue - playing $1 ($2)";
	    
	    eval { larry_queue_start($1,$2); }; logit "Error: $@" if ($@);    
	  }
	elsif (/^\*\*\* The track queue is empty$/)
	  {
	    logit "Debug: start of queue - empty";
	    
	    eval { larry_queue_empty(); }; logit "Error: $@" if ($@);	    
	  }
	elsif (/^\*\*\* ([0-9]+)\. ([!-~]+) queued by ([!-~]+)/)
	  {
	    logit "Debug: queued track $1 is $2";
	    
	    eval { larry_queue($1, $2, $3); }; logit "Error: $@" if ($@);
	  }
	elsif (/^\*\*\* Current volume level is ([0-9]+)\%$/)
	  {
	    logit "Debug: volume level is $1";
	    
	    eval { larry_volume($1); }; logit "Error: $@" if ($@);	    
	  }

	# Event parsing

	elsif (/^--- Playing track \"(.*)\"$/)
	  {
	    logit "Debug: playing $1";
	    
	    eval { larry_playing($1); }; logit "Error: $@" if ($@);	    
	  }
	elsif (/^--- ([!-~]+) has queued track \"(.*)\"$/)
	  {
	    logit "Debug: $1 has queued $2";
	    
	    eval { larry_queued($2, $1); }; logit "Error: $@" if ($@);	    
	  }
	elsif (/^--- ([!-~]+) has deleted track (.*)$/)
	  {
	    logit "Debug: $1 has deleted $2";
	    
	    eval { larry_deleted($2, $1); }; logit "Error: $@" if ($@);
	  }
	elsif (/^--- ([!-~]+) has stopped the current track$/)
	  {
	    logit "Debug: $1 has stopped the current track";
	    
	    eval { larry_stopped(); }; logit "Error: $@" if ($@);	    
	  }
	elsif (/^--- The track queue is empty$/)
	  {
	    logit "Debug: Track queue empty";
	    
	    eval { larry_emptied(); }; logit "Error: $@" if ($@);	    
	  }
	elsif (/^--- Playing finished$/)
	  {
	    logit "Debug: Nothing playing";
	    
	    eval { larry_finished(); }; logit "Error: $@" if ($@);	    
	  }
	elsif (/^--- ([!-~]+) has set the volume to ([0-9]+)\%$/)
	  {
	    logit "Debug: volume now $2";
	    
	    eval { larry_volume($2); }; logit "Error: $@" if ($@);	    
	  }
	elsif (/^--- ([!-~]+)(\s\(.*\)\s|\s)has connected$/)
	  {
	    logit "Debug: $1 ($2) has connected";
	    
	    eval { larry_connected($1, $2); }; logit "Error: $@" if ($@);
	  }
	elsif (/^--- ([!-~]+) is kicked/ || 
	       /^--- ([!-~]+) has quit$/ || 
	       /^--- Connection lost to ([!-~]+)$/)
	  {
	    logit "Debug: $1 has disconnected";
	    
	    eval { larry_disconn($1); }; logit "Error: $@" if ($@);	    
	  }
	elsif (/^--- ([!-~]+) has changed into ([!-~]+)$/)
	  {
	    logit "Debug: $1 renamed to $2";
	    
	    eval { larry_renamed($1, $2); }; logit "Error: $@" if ($@);
	  }
	elsif (/^\*\*\* Playing (.*) is banned$/)
	  {
	    logit "Debug: $1 is banned";

	    eval { larry_banned($1); }; logit "Error: $@" if ($@);
	  }
#	else
#	  {
#	    logit "Blahhh: $_\n";
#	  }

	flush $sock;
	}
  }

# Initialisation sequence

logit "Larry: startup";
signon();
logit "Larry: signed on";

logit "Larry: getting server information";

# Send some commands that will give us a good idea of server status
print $sock ": stirs\n";
print $sock ".colour off\n";
print $sock ".pretty off\n";
print $sock ".list\n";
print $sock ".who\n";
print $sock ".queue\n";
print $sock ".vol\n";

print $sock ": points out that he is a bot\n";
flush $sock;
print $sock "\n";

logit "Larry: listening for activity";

mainloop();

logit "Larry: Disconnected";

sleep 1;

